package module

import (
	gocontext "context"
	"database/sql"
	"errors"
	"fmt"
	"math"
	"runtime/debug"
	"strconv"
	"strings"
	"sync"
	"time"

	"github.com/huandu/go-sqlbuilder"
	"gitlab.alibaba-inc.com/pai_biz_arch/pairec/context"
	"gitlab.alibaba-inc.com/pai_biz_arch/pairec/log"
	"gitlab.alibaba-inc.com/pai_biz_arch/pairec/persist/holo"
	"gitlab.alibaba-inc.com/pai_biz_arch/pairec/recconf"
	"gitlab.alibaba-inc.com/pai_biz_arch/pairec/utils"
	"gitlab.alibaba-inc.com/pai_biz_arch/pairec/utils/sqlutil"
)

type FeatureHologresDao struct {
	*FeatureBaseDao
	db                   *sql.DB
	table                string
	userFeatureKeyName   string
	itemFeatureKeyName   string
	userSelectFields     string
	itemSelectFields     string
	sequenceOfflineTable string
	mu                   sync.RWMutex
	userStmt             *sql.Stmt
	itemStmtMap          map[int]*sql.Stmt
	userSequenceStmtMap  map[string]*sql.Stmt
	onlineSequenceStmt   *sql.Stmt
	offlineSequenceStmt  *sql.Stmt
}

func NewFeatureHologresDao(config recconf.FeatureDaoConfig) *FeatureHologresDao {
	dao := &FeatureHologresDao{
		FeatureBaseDao:       NewFeatureBaseDao(&config),
		table:                config.HologresTableName,
		userFeatureKeyName:   config.UserFeatureKeyName,
		itemFeatureKeyName:   config.ItemFeatureKeyName,
		userSelectFields:     config.UserSelectFields,
		itemSelectFields:     config.ItemSelectFields,
		sequenceOfflineTable: config.SequenceOfflineTableName,
		itemStmtMap:          make(map[int]*sql.Stmt),
		userSequenceStmtMap:  make(map[string]*sql.Stmt),
	}
	hologres, err := holo.GetPostgres(config.HologresName)
	if err != nil {
		log.Error(fmt.Sprintf("error=%v", err))
		return nil
	}
	dao.db = hologres.DB
	return dao
}
func (d *FeatureHologresDao) getItemStmt(key int) *sql.Stmt {
	d.mu.RLock()
	defer d.mu.RUnlock()
	return d.itemStmtMap[key]
}
func (d *FeatureHologresDao) getUserSequenceStmt(key string) *sql.Stmt {
	d.mu.RLock()
	defer d.mu.RUnlock()
	return d.userSequenceStmtMap[key]
}
func (d *FeatureHologresDao) FeatureFetch(user *User, items []*Item, context *context.RecommendContext) {
	if d.featureStore == Feature_Store_User && d.featureType == Feature_Type_Sequence {
		d.userSequenceFeatureFetch(user, context)
	} else if d.featureStore == Feature_Store_User {
		d.userFeatureFetch(user, context)
	} else {
		d.itemsFeatureFetch(items, context)
	}
}

func (d *FeatureHologresDao) userFeatureFetch(user *User, context *context.RecommendContext) {
	defer func() {
		if err := recover(); err != nil {
			log.Error(fmt.Sprintf("requestId=%s\tmodule=FeatureHologresDao\terror=%v", context.RecommendId, err))
			return
		}
	}()

	comms := strings.Split(d.featureKey, ":")
	if len(comms) < 2 {
		log.Error(fmt.Sprintf("requestId=%s\tuid=%s\terror=featureKey error(%s)", context.RecommendId, user.Id, d.featureKey))
		return
	}

	key := user.StringProperty(comms[1])
	if key == "" {
		log.Error(fmt.Sprintf("requestId=%s\tmodule=FeatureHologresDao\terror=property not found(%s)", context.RecommendId, comms[1]))
		return
	}

	builder := sqlbuilder.PostgreSQL.NewSelectBuilder()
	builder.Select(d.userSelectFields)
	builder.From(d.table)
	builder.Where(builder.Equal(d.userFeatureKeyName, key))

	sqlquery, args := builder.Build()
	if d.userStmt == nil {
		d.mu.Lock()
		if d.userStmt == nil {
			stmt, err := d.db.Prepare(sqlquery)
			if err != nil {
				log.Error(fmt.Sprintf("requestId=%s\tmodule=FeatureHologresDao\terror=hologres error(%v)", context.RecommendId, err))
				d.mu.Unlock()
				return
			}
			d.userStmt = stmt
			d.mu.Unlock()
		} else {
			d.mu.Unlock()
		}
	}

	ctx, cancel := gocontext.WithTimeout(gocontext.Background(), 100*time.Millisecond)
	defer cancel()
	rows, err := d.userStmt.QueryContext(ctx, args...)
	if err != nil {
		if errors.Is(err, gocontext.DeadlineExceeded) {
			log.Warning("module=FeatureHologresDao\tevent=userFeatureFetch\ttimeout=100")
			return
		}
		log.Error(fmt.Sprintf("requestId=%s\tmodule=FeatureHologresDao\terror=hologres error(%v)", context.RecommendId, err))
		return
	}

	defer rows.Close()
	columns, err := rows.ColumnTypes()
	if err != nil {
		log.Error(fmt.Sprintf("requestId=%s\tmodule=FeatureHologresDao\terror=hologres error(%v)", context.RecommendId, err))
		return
	}
	values := sqlutil.ColumnValues(columns)

	for rows.Next() {
		if err := rows.Scan(values...); err == nil {
			for i, column := range columns {
				name := column.Name()
				if name == d.userFeatureKeyName {
					continue
				}

				val := values[i]
				switch v := val.(type) {
				case *sql.NullString:
					if v.Valid {
						user.AddProperty(name, v.String)
					}
				case *sql.NullInt32:
					if v.Valid {
						user.AddProperty(name, v.Int32)
					}
				case *sql.NullInt64:
					if v.Valid {
						user.AddProperty(name, v.Int64)
					}
				case *sql.NullFloat64:
					if v.Valid {
						user.AddProperty(name, v.Float64)
					}
				default:
				}
			}
		} else {
			log.Error(fmt.Sprintf("requestId=%s\tmodule=FeatureHologresDao\terror=hologres error(%v)", context.RecommendId, err))
		}
	}
}

type sequenceInfo struct {
	itemId          string
	event           string
	playTime        float64
	timestamp       int64
	dimensionFields []sql.NullString
}

func (d *FeatureHologresDao) userSequenceFeatureFetch(user *User, context *context.RecommendContext) {
	defer func() {
		if err := recover(); err != nil {
			stack := string(debug.Stack())
			log.Error(fmt.Sprintf("requestId=%s\tmodule=FeatureHologresDao\terror=%v\tstack=%v", context.RecommendId, err, strings.ReplaceAll(stack, "\n", "\t")))
			return
		}
	}()

	comms := strings.Split(d.featureKey, ":")
	if len(comms) < 2 {
		log.Error(fmt.Sprintf("requestId=%s\tuid=%s\terror=featureKey error(%s)", context.RecommendId, user.Id, d.featureKey))
		return
	}

	key := user.StringProperty(comms[1])
	if key == "" {
		log.Error(fmt.Sprintf("requestId=%s\tmodule=FeatureHologresDao\terror=property not found(%s)", context.RecommendId, comms[1]))
		return
	}

	currTime := time.Now().Unix()
	var onlineSequences []*sequenceInfo
	selectFields := []string{"item_id", "event", "play_time", "timestamp"}
	if len(d.sequenceDimFields) > 0 {
		selectFields = append(selectFields, d.sequenceDimFields...)
	}

	onlineFunc := func() {
		builder := sqlbuilder.PostgreSQL.NewSelectBuilder()
		builder.Select(selectFields...)
		builder.From(d.table)
		where := []string{builder.Equal(d.userFeatureKeyName, key), builder.GreaterThan("timestamp", currTime-86400*5)}
		if d.sequenceEvent != "" {
			where = append(where, builder.Equal("event", d.sequenceEvent))
		}
		builder.Where(where...).Limit(d.sequenceLength)
		builder.OrderBy("timestamp").Desc()

		sqlquery, args := builder.Build()
		if d.onlineSequenceStmt == nil {
			d.mu.Lock()
			if d.onlineSequenceStmt == nil {
				stmt, err := d.db.Prepare(sqlquery)
				if err != nil {
					log.Error(fmt.Sprintf("requestId=%s\tmodule=FeatureHologresDao\terror=hologres error(%v)", context.RecommendId, err))
					d.mu.Unlock()
					return
				}
				d.onlineSequenceStmt = stmt
				d.mu.Unlock()
			} else {
				d.mu.Unlock()
			}
		}
		ctx, cancel := gocontext.WithTimeout(gocontext.Background(), 100*time.Millisecond)
		defer cancel()
		rows, err := d.onlineSequenceStmt.QueryContext(ctx, args...)
		if err != nil {
			if errors.Is(err, gocontext.DeadlineExceeded) {
				log.Warning("module=FeatureHologresDao\tevent=userSequenceFeatureFetch\ttimeout=100")
				return
			}
			log.Error(fmt.Sprintf("requestId=%s\tmodule=FeatureHologresDao\terror=hologres error(%v)", context.RecommendId, err))
			return
		}

		defer rows.Close()
		for rows.Next() {
			seq := new(sequenceInfo)
			dst := []interface{}{&seq.itemId, &seq.event, &seq.playTime, &seq.timestamp}
			if len(d.sequenceDimFields) > 0 {
				seq.dimensionFields = make([]sql.NullString, len(d.sequenceDimFields))
				for i := range seq.dimensionFields {
					dst = append(dst, &seq.dimensionFields[i])
				}
			}
			if err := rows.Scan(dst...); err == nil {
				if t, exist := d.sequencePlayTimeMap[seq.event]; exist {
					if seq.playTime <= t {
						continue
					}
				}
				onlineSequences = append(onlineSequences, seq)
			} else {
				log.Error(fmt.Sprintf("requestId=%s\tmodule=FeatureHologresDao\terror=hologres error(%v)", context.RecommendId, err))
			}
		}

	}

	var offlineSequences []*sequenceInfo
	offlineFunc := func() {
		builder := sqlbuilder.PostgreSQL.NewSelectBuilder()
		builder.Select(selectFields...)
		builder.From(d.sequenceOfflineTable)
		where := []string{builder.Equal(d.userFeatureKeyName, key)}
		if d.sequenceEvent != "" {
			where = append(where, builder.Equal("event", d.sequenceEvent))
		}
		builder.Where(where...).Limit(d.sequenceLength)
		builder.OrderBy("timestamp").Desc()

		sqlquery, args := builder.Build()
		if d.offlineSequenceStmt == nil {
			d.mu.Lock()
			if d.offlineSequenceStmt == nil {
				stmt, err := d.db.Prepare(sqlquery)
				if err != nil {
					log.Error(fmt.Sprintf("requestId=%s\tmodule=FeatureHologresDao\terror=hologres error(%v)", context.RecommendId, err))
					d.mu.Unlock()
					return
				}
				d.offlineSequenceStmt = stmt
				d.mu.Unlock()
			} else {
				d.mu.Unlock()
			}
		}
		ctx, cancel := gocontext.WithTimeout(gocontext.Background(), 100*time.Millisecond)
		defer cancel()
		rows, err := d.offlineSequenceStmt.QueryContext(ctx, args...)
		if err != nil {
			if errors.Is(err, gocontext.DeadlineExceeded) {
				log.Warning("module=FeatureHologresDao\tevent=userSequenceFeatureFetch\ttimeout=100")
				return
			}
			log.Error(fmt.Sprintf("requestId=%s\tmodule=FeatureHologresDao\terror=hologres error(%v)", context.RecommendId, err))
			return
		}

		defer rows.Close()
		for rows.Next() {
			seq := new(sequenceInfo)
			dst := []interface{}{&seq.itemId, &seq.event, &seq.playTime, &seq.timestamp}
			if len(d.sequenceDimFields) > 0 {
				seq.dimensionFields = make([]sql.NullString, len(d.sequenceDimFields))
				for i := range seq.dimensionFields {
					dst = append(dst, &seq.dimensionFields[i])
				}
			}
			if err := rows.Scan(dst...); err == nil {
				if t, exist := d.sequencePlayTimeMap[seq.event]; exist {
					if seq.playTime <= t {
						continue
					}
				}
				offlineSequences = append(offlineSequences, seq)
			} else {
				log.Error(fmt.Sprintf("requestId=%s\tmodule=FeatureHologresDao\terror=hologres error(%v)", context.RecommendId, err))
			}
		}

	}

	var wg sync.WaitGroup

	wg.Add(1)
	go func() {
		defer wg.Done()
		onlineFunc()
	}()
	if d.sequenceOfflineTable != "" {
		wg.Add(1)
		go func() {
			defer wg.Done()
			offlineFunc()
		}()
	}
	wg.Wait()

	if len(offlineSequences) > 0 {
		index := 0
		for index < len(onlineSequences) {
			if onlineSequences[index].timestamp < offlineSequences[0].timestamp {
				break
			}
			index++
		}

		onlineSequences = onlineSequences[:index]
		onlineSequences = append(onlineSequences, offlineSequences...)
		if len(onlineSequences) > d.sequenceLength {
			onlineSequences = onlineSequences[:d.sequenceLength]
		}
	}

	var sequences []string
	sequenceMap := make(map[string]bool, 0)

	for _, seq := range onlineSequences {
		key := fmt.Sprintf("%s#%s", seq.itemId, seq.event)
		if _, exist := sequenceMap[key]; !exist {
			sequenceMap[key] = true
			seqStr := fmt.Sprintf("%s#%d#%s#%.2f#%d", seq.itemId, seq.timestamp, seq.event, seq.playTime, currTime-seq.timestamp)
			for _, field := range seq.dimensionFields {
				if field.Valid {
					seqStr += "#" + field.String
				} else {
					seqStr += "#"
				}
			}
			sequences = append(sequences, seqStr)
		}
	}

	delim := d.sequenceDelim
	if delim == "" {
		delim = ";"
	}
	user.AddProperty(d.sequenceName, strings.Join(sequences, delim))
}

func (d *FeatureHologresDao) itemsFeatureFetch(items []*Item, context *context.RecommendContext) {
	defer func() {
		if err := recover(); err != nil {
			log.Error(fmt.Sprintf("requestId=%s\tmodule=FeatureHologresDao\terror=%v", context.RecommendId, err))
			return
		}
	}()

	fk := d.featureKey
	if fk != "item:id" {
		comms := strings.Split(d.featureKey, ":")
		if len(comms) < 2 {
			log.Error(fmt.Sprintf("requestId=%s\tevent=itemsFeatureFetch\terror=featureKey error(%s)", context.RecommendId, d.featureKey))
			return
		}

		fk = comms[1]
	}

	cpuCount := utils.MaxInt(int(math.Ceil(float64(len(items))/float64(1000))), 1)
	requestCh := make(chan []*Item, cpuCount)
	defer close(requestCh)

	if cpuCount == 1 {
		requestCh <- items
	} else {
		maps := make(map[int][]*Item)
		for i, item := range items {
			maps[i%cpuCount] = append(maps[i%cpuCount], item)
		}

		for _, itemlist := range maps {
			requestCh <- itemlist
		}

	}

	var wg sync.WaitGroup
	for i := 0; i < cpuCount; i++ {
		wg.Add(1)
		go func() {
			defer wg.Done()
			select {
			case itemlist := <-requestCh:
				var keys []interface{}
				key2Item := make(map[string]*Item, len(itemlist))
				for _, item := range itemlist {
					var key string
					if fk == "item:id" {
						key = string(item.Id)
					} else {
						key = item.StringProperty(fk)
					}
					keys = append(keys, key)
					key2Item[key] = item
				}

				builder := sqlbuilder.PostgreSQL.NewSelectBuilder()
				builder.Select(d.itemSelectFields)
				builder.From(d.table)
				builder.Where(builder.In(d.itemFeatureKeyName, keys...))

				sqlquery, args := builder.Build()
				stmtkey := len(keys)
				stmt := d.getItemStmt(stmtkey)
				if stmt == nil {
					d.mu.Lock()
					stmt = d.itemStmtMap[stmtkey]
					if stmt == nil {
						stmt2, err := d.db.Prepare(sqlquery)
						if err != nil {
							log.Error(fmt.Sprintf("requestId=%s\tmodule=FeatureHologresDao\terror=hologres error(%v)", context.RecommendId, err))
							d.mu.Unlock()
							return
						}
						d.itemStmtMap[stmtkey] = stmt2
						stmt = stmt2
						d.mu.Unlock()
					} else {
						d.mu.Unlock()
					}
				}

				rowsChannel := make(chan *sql.Rows, 1)
				ctx, cancel := gocontext.WithTimeout(gocontext.Background(), 100*time.Millisecond)
				defer cancel()
				// async invoke sql query
				go func() {
					rows, err := stmt.Query(args...)
					if err != nil {
						rowsChannel <- nil
						log.Error(fmt.Sprintf("requestId=%s\tmodule=FeatureHologresDao\terror=hologres error(%v)", context.RecommendId, err))
						return
					}

					// check query is timeout
					select {
					case <-ctx.Done():
						if rows != nil {
							rows.Close()
						}
						log.Error(fmt.Sprintf("requestId=%s\tmodule=FeatureHologresDao\terror=%v", context.RecommendId, ctx.Err()))
						return
					default:
					}
					rowsChannel <- rows
					return
				}()

				var rows *sql.Rows
				select {
				case <-ctx.Done():
					log.Error(fmt.Sprintf("requestId=%s\tmodule=FeatureHologresDao\terror=hologres error(%v)", context.RecommendId, ctx.Err()))
					return
				case rows = <-rowsChannel:
					if rows == nil {
						return
					}
				}

				defer rows.Close()
				columns, err := rows.ColumnTypes()
				if err != nil {
					log.Error(fmt.Sprintf("requestId=%s\tmodule=FeatureHologresDao\terror=hologres error(%v)", context.RecommendId, err))
					return
				}

				values := sqlutil.ColumnValues(columns)
				for rows.Next() {
					if err := rows.Scan(values...); err == nil {
						var item *Item
						properties := make(map[string]interface{}, len(values))
						for i, column := range columns {
							name := column.Name()
							val := values[i]
							if i == 0 {
								var key string
								switch v := val.(type) {
								case *sql.NullString:
									if v.Valid {
										key = v.String
									}
								case *sql.NullInt32:
									if v.Valid {
										key = strconv.Itoa(int(v.Int32))
									}
								case *sql.NullInt64:
									if v.Valid {
										key = strconv.Itoa(int(v.Int64))
									}
								}
								if key == "" {
									break
								}
								item = key2Item[key]
								continue
							}

							switch v := val.(type) {
							case *sql.NullString:
								if v.Valid {
									properties[name] = v.String
								}
							case *sql.NullInt32:
								if v.Valid {
									properties[name] = v.Int32
								}
							case *sql.NullInt64:
								if v.Valid {
									properties[name] = v.Int64
								}
							case *sql.NullFloat64:
								if v.Valid {
									properties[name] = v.Float64
								}
							default:
							}
						}
						if nil != item && len(properties) > 0 {
							item.AddProperties(properties)
						}

					} else {
						log.Error(fmt.Sprintf("requestId=%s\tmodule=FeatureHologresDao\terror=hologres error(%v)", context.RecommendId, err))
					}
				}
			default:
			}
		}()
	}
	wg.Wait()
}
